
/*
 * polldevice.{cc,hh} -- element steals packets from Linux devices by polling.
 * 
 * Benjie Chen
 *
 * Copyright (c) 1999 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. For more information, see the `COPYRIGHT' file in the source
 * distribution.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include "glue.hh"
#include "polldevice.hh"
#include "error.hh"
#include "packet.hh"
#include "confparse.hh"
#include "router.hh"

extern "C" {
#include <linux/netdevice.h>
}


PollDevice::PollDevice()
  : _dev(0)
{
  add_output();
}

PollDevice::PollDevice(const String &devname)
  : _devname(devname), _dev(0)
{
#ifdef CLICK_BENCHMARK
  _bm_done = 0;
#endif
  add_output();
}

PollDevice::~PollDevice()
{
  uninitialize();
}

void
PollDevice::static_initialize()
{
}

void
PollDevice::static_cleanup()
{
}


PollDevice *
PollDevice::clone() const
{
  return new PollDevice();
}

int
PollDevice::configure(const String &conf, ErrorHandler *errh)
{
  return cp_va_parse(conf, this, errh,
		     cpString, "interface name", &_devname,
		     cpEnd);
}


/*
 * Use Linux interface for polling, added by us, in include/linux/netdevice.h,
 * to poll devices.
 */
int
PollDevice::initialize(ErrorHandler *errh)
{
  _dev = dev_get(_devname.cc());
  if (!_dev)
    return errh->error("no device `%s'", _devname.cc());
  if (!_dev->pollable) 
    return errh->error("device `%s' not pollable", _devname.cc());
  _dev->intr_off(_dev);
  _dev->intr_defer = 1;
  _idle = 0;
  _total_intr_wait = 0;
  return 0;
}


void
PollDevice::uninitialize()
{
  if (_dev)
  { 
    _dev->intr_defer = 0; 
    _dev->intr_on(_dev);
    click_chatter
	("PollDevice(%s): waited with intr on %d times\n", 
	 _dev->name, _total_intr_wait);
  }
}


Packet *
PollDevice::pull(int)
{
  extern int rtm_ipackets;
  extern unsigned long long rtm_ibytes;
  struct sk_buff *skb;

  skb = _dev->rx_poll(_dev);
  _dev->fill_rx(_dev);
  _dev->clean_tx(_dev);

  if (skb != 0L)
  {
    _idle = 0;
    rtm_ipackets++;
    rtm_ibytes += skb->len;

    assert(skb->data - skb->head >= 14);
    assert(skb->mac.raw == skb->data - 14);
    assert(skb_shared(skb) == 0); /* else skb = skb_clone(skb, GFP_ATOMIC); */

    /* Retrieve the ether header. */
    skb_push(skb, 14);

    Packet *p = Packet::make(skb);
    if(skb->pkt_type == PACKET_MULTICAST || skb->pkt_type == PACKET_BROADCAST)
      p->set_mac_broadcast_anno(1);

    return p;
  } 
  else 
  {
    _idle++;
    return 0L;
  }
}
 

bool
PollDevice::still_busy() const
{
  return _idle <= 100;
}

void
PollDevice::set_wakeup_when_busy()
{
  /* put self on intr wait queue */
  _self_wq.task = current;
  _self_wq.next = NULL;
  add_wait_queue(&(_dev->intr_wq), &_self_wq);

  /* turn interrupts back on */
  _total_intr_wait++;
  _dev->intr_on(_dev);
}

void
PollDevice::woke_up()
{
  _dev->intr_off(_dev);
  remove_wait_queue(&(_dev->intr_wq), &_self_wq);
}

ELEMENT_REQUIRES(false)
EXPORT_ELEMENT(PollDevice)
